#include <iostream>
#include <string>
#include <list>

// #include "OPTION.h"

#include "Option.h"
#include "Gauss.h"
#include "FeaturePointTable.h"
#include "TriVecDescTable.h"
#include "Vote.h"
#include "CrsPoint.h"
#include "fundamental_estimation.h"

double EPS     = 1e-10;
double DET_EPS = 1e-10;

using std::cout;
using std::cerr;
using std::endl;
using std::list;
using std::string;

int main(int argc, char *argv[])
{
   // analize option
//   OPTION option;
//   option.analysis(argc, argv);
   Option option;
   try{
      option.Parse(argc,argv);
   }
   catch(OptionException){
      return 1;
   }

   CImage       cimg1, cimg2;
   Image        img1, img2;
   Gauss        gauss(option.filter_range(), option.var());
   FeaturePointTable    feature1,feature2;
   TriVectorTable       tv_table1,tv_table2;
   list<CORRES_POINT>   crsp_list;
   Vote         vote;
   int          num;

   string       img1name(option.imgfn1());
   string       img2name(option.imgfn2());
   size_t       strsize;

   // load 1st image
   strsize = img1name.size();
   if (img1name.rfind(".ppm") < strsize)
   {
      cimg1.load(img1name.c_str());           // load ppm image
      cimg1.conv_gray(img1);             // convert to gray image
      if (option.is_set_verbose())
         cerr << "** Convert color images to graylevels." << endl;
   }
   else if (img1name.rfind(".pgm") < strsize)
   {
      img1.load(img1name.c_str());           // load pgm image
      if (option.is_set_verbose())
         cerr << "** Load the first image file." << endl;
   }

   // load 2nd image
   strsize = img2name.size();
   if (img2name.rfind(".ppm") < strsize)
   {
      cimg2.load(img2name.c_str());           // load ppm image
      cimg2.conv_gray(img2);             // convert to gray image
      if (option.is_set_verbose())
         cerr << "** Convert color images to graylevels." << endl;
   }
   else if (img2name.rfind(".pgm") < strsize)
   {
      // load pgm image
      img2.load(img2name.c_str());            // load pgm image
      if (option.is_set_verbose())
         cerr << "** Load the first image file." << endl;
   }

   // feature points extraction
   if (option.is_set_f1_name() && option.is_set_f2_name())
   {
      if (option.is_set_verbose())
         cerr << "** Read feature point data from files: "
              << option.f1_name().c_str() << ", "
              << option.f2_name().c_str() << endl;
      feature1.input_from_file(option.f1_name().c_str() );
      feature2.input_from_file(option.f2_name().c_str() ); 
   }
   else
   {
      if (option.is_set_verbose())
         cerr << "** Apply Harris operater to the images: "
              << option.imgfn1().c_str() << ", "
              << option.imgfn2().c_str() << endl;

      feature1.extract_from_image(img1, gauss.range, gauss.var,
                                  option.rk(), option.harris_th(),
                                  option.Pmax(),
                                  option.harris_d_min() );
      feature2.extract_from_image(img2, gauss.range, gauss.var,
                                  option.rk(), option.harris_th(),
                                  option.Pmax(),
                                  option.harris_d_min() );
   }
   // compute covariance matrices
   if (option.is_set_verbose())
      cerr << "** Compute the covariance matrices of the feature points."
           << endl;

   double normalize_factor1 =
      feature1.calc_ellipse_parameter(img1, gauss,
                                      option.is_set_use_CC(),
                                      option.window_range() );
   double normalize_factor2 = 
      feature2.calc_ellipse_parameter(img2, gauss,
                                      option.is_set_use_CC(),
                                      option.window_range() );
   if (option.is_set_verbose())
   {
      cerr << "** Image 1:\n"
           << "**** The number of feature points:\t"
           << feature1.size() << "\n"
           << "**** Scale factor for normalization:\t"
           << normalize_factor1 << "\n"
           << "** Image 2:\n"
           << "**** The number of feature points:\t"
           << feature2.size() << "\n"
           << "**** Scale factor for normalization:\t"
           << normalize_factor2 << endl;
   }
   
   // normalize the sizes of covariance matrices
   feature1.normalize(normalize_factor1);
   feature2.normalize(normalize_factor1);  
   if (option.is_set_verbose())
      cerr << "** Construct TVD Table 1:\t";

   // construct triplet vector descriptor
   tv_table1.make_TriVector_table(img1, feature1, gauss,
                                  option.dim(), option.d_min(), option.d_max(),
                                  option.radius_div(), option.radius_scale(),
                                  option.angle_div(),
                                  option.is_set_use_CC());
   if (option.is_set_verbose())
   {
      cerr << tv_table1.count_TriVector() << endl;
      cerr << "** Construct TVD Table 2:\t";
   }
   tv_table2.make_TriVector_table(img2, feature2, gauss, 
                                  option.dim(), option.d_min(), option.d_max(),
                                  option.radius_div(), option.radius_scale(),
                                  option.angle_div(),
                                  option.is_set_use_Derivative());
   if (option.is_set_verbose())
   {
      cerr << tv_table2.count_TriVector() << endl;
  
      cerr << "** Initialize the voting table." << endl;
   }  
   // vote TVDs and Triangle Regions
   vote.initialize(option.pv_th(), option.str_th(), option.triangle_th(),
                   option.pv_weight(), option.str_weight(),
                   option.triangle_weight(),
                   feature1.size(), feature2.size());
  
  
   if (option.is_set_verbose())
      cerr << "** Start voting." << endl;

   vote.voting(tv_table1, tv_table2, feature1, feature2, img1, img2,
               gauss, option.triangle_length(), option.vote_th(), crsp_list);
   if(!option.is_set_no_use_RANSAC()){
      if(option.is_set_verbose()){
         cerr << "** Start RANSAC."<<endl;
      }
      // Fundamental matrix estimation using RANSAC
      fundamental_estimation(crsp_list, img1.xl(),
                             img1.yl(), option.focal_length(),
                             option.distortion());
   }
   if(option.is_set_verbose()){
      cerr << "** Output correspondences." << endl;
   }
   // output correspondences to stdout
   num = show_match_point(crsp_list);
   cerr << "The total number of correspondences: " << num << endl;
  
   return 0;
}
