//
// Improved Harris Corner Detector
//

#include	<cmath>

#include	"Image.h"
#include	"Harris.h"
#include	"HarrisTools.h"

#define		MaxGray		255

// Harris corner detector w/o output image
int
ImpHarris(const Image &InImg,
          int **up0, int **vp0, double **value,
          int *cnr_num, int Ns, double sigma, double rk,
          double s, int max_corner, double min_dist)
{
   double	r, d;
   double       Ix, Iy;
   double	Ix2, Iy2, Ixy;
   double       gx, gy;
   double	gxx, gyy, gxy;
   double	max_r, r_thd;
   int		corner_num;
   int		x, x0, xx, u0, v0;
   int		offwidth, offheight;
   int		uoff, voff;
   int		dblw;
   
   int		*up = new int [max_corner];
   int		*vp = new int [max_corner];

   int		width = InImg.xl();
   int		height = InImg.yl();
   const int    *Im = InImg.data;

   double	*Gw   = new double [(2*Ns+1) * (2*Ns+1)];
   float	*Imx  = new float [width * height];
   float	*Imy  = new float [width * height];
   float	*Imxx = new float [width * height];
   float	*Imyy = new float [width * height];
   float	*Imxy = new float [width * height];
   float	*out  = new float [width * height];
   double	*dp = new double [max_corner];

   // memory allocation check
   if (!up || !vp || !Gw || !Imx || !Imy || !Imxx || !Imyy || !Imxy ||
       !out || !dp)
      return 0;

   (*up0) = up;
   (*vp0) = vp;
   
   make_Gauss_window(Gw, Ns, sigma);

   dblw = width * 2;

   voff = uoff = 2;
   offwidth = width - uoff;
   offheight = height - voff;
   
   for (v0 = voff; v0 < offheight; v0++)
   {
      x = (v0 * width) + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 Ix = G2 * (Im[x + 2] - Im[x - 2])
            + G1 * (Im[x + 1] - Im[x - 1]);
	 Iy = G2 * (Im[x + dblw] - Im[x - dblw])
	    + G1 * (Im[x + width] - Im[x - width]);
	 
	 Imx[x] = Ix;
	 Imy[x] = Iy;
      }
   }

   uoff += 2;
   voff += 2;
   offwidth -= 2;
   offheight -= 2;

   for (v0 = voff; v0 < offheight; v0++)
   {
      x = (v0 * width) + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 gx = Imx[x - 2] + Imx[x - 1] + Imx[x] + Imx[x + 1] + Imx[x + 2];
	 gy = Imy[x - dblw] + Imy[x - width] + Imy[x]
	    + Imy[x + width] + Imy[x + dblw];
	 Imxx[x] = gx * gx;
	 Imyy[x] = gy * gy;
	 Imxy[x] = gx * gy;
      }
   }

   uoff += Ns;
   voff += Ns;
   offwidth -= Ns;
   offheight -= Ns;
   max_r = 0.0;
   
   for (v0 = voff; v0 < offheight; v0++)
   {
      x = v0 * width + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 gxx = gyy = gxy = 0.0;
	 xx = 0;
	 for (int vv = -Ns; vv <= Ns; vv++)
	 {
	    x0 = (v0 + vv) * width + u0;
	    for (int uu = -Ns; uu <= Ns; uu++)
	    {
	       gxx += Gw[xx] * Imxx[x0 + uu];
	       gyy += Gw[xx] * Imyy[x0 + uu];
	       gxy += Gw[xx] * Imxy[x0 + uu];
	       xx++;
	    }
	 }
	 
	 Ix2 = gxx * gxx;
	 Iy2 = gyy * gyy;
	 Ixy = gxy * gxy;
	 r = (Ix2 * Iy2 - sqr(Ixy)) - rk * sqr(Ix2 + Iy2);

	 if (r > max_r)
	    max_r = r;
	 
	 out[x] = r;
      }
   }

   r_thd = max_r * s;

   uoff += 2;
   voff += 2;
   offwidth -= 2;
   offheight -= 2;

   corner_num = 0;
   for (v0 = voff; v0 < offheight; v0++)
   {
      x = v0 * width + uoff;
      for (u0 = uoff; u0 < offwidth; u0++, x++)
      {
	 d = out[x];
	 if (d <= r_thd)  continue;

	 if (   check_ge_upperleft(d, out, x, width)
             && check_gt_lowerright(d, out, x, width))
	 {
	    corner_num = add_list(u0, v0, d, up, vp, dp,
				  corner_num, max_corner, min_dist);
	 }
      }
   }

   // if you want, Harris values are copied.
   if (value)
   {
      (*value) = new double [corner_num];
      if(!(*value ))
         return 0;
      for(int i=0; i < corner_num; i++)
         (*value)[i] = dp[i];
   }
	
// delete arrays
   delete [] dp;
   delete [] out;
   delete [] Imxy;
   delete [] Imyy;
   delete [] Imxx;
   delete [] Imy;
   delete [] Imx;
   delete [] Gw;
   
   return (*cnr_num = corner_num);
}
